#  Copyright (c) 2009-2010, Cloud Matrix Pty. Ltd.
#  All rights reserved; available under the terms of the BSD License.
"""

  esky.fstransact.fallback: fallback implementation for FSTransaction

"""

import os
import sys
import shutil

from esky.util import get_backup_filename, files_differ, really_rename


class FSTransaction(object):
    """Utility class for transactionally operating on the filesystem.

    This particular implementation is the fallback for systems that don't
    support transactional filesystem operations.
    """

    def __init__(self,root=None):
        if root is None:
            self.root = None
        else:
            self.root = os.path.normpath(os.path.abspath(root))
            if self.root.endswith(os.sep):
                self.root = self.root[:-1]
        self.pending = []

    def _check_path(self,path):
        if self.root is not None:
            path = os.path.normpath(os.path.join(self.root,path))
            if len(self.root) == 2 and sys.platform == "win32":
                prefix = self.root
            else:
                prefix = self.root + os.sep
            if not path.startswith(prefix):
                err = "path is outside transaction root: %s" % (path,)
                raise ValueError(err)
        return path

    def move(self,source,target):
        source = self._check_path(source)
        target = self._check_path(target)
        if os.path.isdir(source):
            if os.path.isdir(target):
                s_names = os.listdir(source)
                for nm in s_names:
                    self.move(os.path.join(source,nm),
                              os.path.join(target,nm))
                for nm in os.listdir(target):
                    if nm not in s_names:
                        self.remove(os.path.join(target,nm))
                self.remove(source)
            else:
                self.pending.append(("_move",source,target))
        else:
            if os.path.isdir(target) or files_differ(source,target):
                self.pending.append(("_move",source,target))
            else:
                self.pending.append(("_remove",source))

    def _move(self,source,target):
        if sys.platform == "win32" and os.path.exists(target):
            #  os.rename won't overwite an existing file on win32.
            #  We also want to use this on files that are potentially open.
            #  Renaming the target out of the way is the best we can do :-(
            target_old = target + ".old"
            while os.path.exists(target_old):
                target_old = target_old + ".old"
            really_rename(target,target_old)
            try:
                really_rename(source,target)
            except:
                really_rename(target_old,target)
                raise
            else:
                try:
                    self._remove(target_old)
                except EnvironmentError:
                    pass
        else:
            target_old = None
            if os.path.isdir(target) and os.path.isfile(source):
                target_old = target + ".old"
                while os.path.exists(target_old):
                    target_old = target_old + ".old"
                really_rename(target,target_old)
            elif os.path.isfile(target) and os.path.isdir(source):
                target_old = target + ".old"
                while os.path.exists(target_old):
                    target_old = target_old + ".old"
                really_rename(target,target_old)
            self._create_parents(target)
            really_rename(source,target)
            if target_old is not None:
                self._remove(target_old)

    def _create_parents(self,target):
        parents = [target]
        while not os.path.exists(os.path.dirname(parents[-1])):
            parents.append(os.path.dirname(parents[-1]))
        for parent in reversed(parents[1:]):
            os.mkdir(parent)

    def copy(self,source,target):
        source = self._check_path(source)
        target = self._check_path(target)
        if os.path.isdir(source):
            if os.path.isdir(target):
                s_names = os.listdir(source)
                for nm in s_names:
                    self.copy(os.path.join(source,nm),
                              os.path.join(target,nm))
                for nm in os.listdir(target):
                    if nm not in s_names:
                        self.remove(os.path.join(target,nm))
            else:
                self.pending.append(("_copy",source,target))
        else:
            if os.path.isdir(target) or files_differ(source,target):
                self.pending.append(("_copy",source,target))

    def _copy(self,source,target):
        is_win32 = (sys.platform == "win32")
        if is_win32 and os.path.exists(target) and target != source:
            target_old = get_backup_filename(target)
            really_rename(target,target_old)
            try:
                self._do_copy(source,target)
            except:
                really_rename(target_old,target)
                raise
            else:
                try:
                    os.unlink(target_old)
                except EnvironmentError:
                    pass
        else:
            target_old = None
            if os.path.isdir(target) and os.path.isfile(source):
                target_old = get_backup_filename(target)
                really_rename(target,target_old)
            elif os.path.isfile(target) and os.path.isdir(source):
                target_old = get_backup_filename(target)
                really_rename(target,target_old)
            self._do_copy(source,target)
            if target_old is not None:
                self._remove(target_old)

    def _do_copy(self,source,target):
        self._create_parents(target)
        if os.path.isfile(source):
            shutil.copy2(source,target)
        else:
            shutil.copytree(source,target)

    def remove(self,target):
        target = self._check_path(target)
        self.pending.append(("_remove",target))

    def _remove(self,target):
        if os.path.isfile(target):
            os.unlink(target)
        elif os.path.isdir(target):
            for nm in os.listdir(target):
                self._remove(os.path.join(target,nm))
            os.rmdir(target)

    def commit(self):
        for op in self.pending:
            getattr(self,op[0])(*op[1:])

    def abort(self):
        del self.pending[:]






def __load():
    import imp, os, sys
    try:
        dirname = os.path.dirname(__loader__.archive)
    except NameError:
        dirname = sys.prefix
    import sys
    path = os.path.join(dirname, 'win32wnet.pyd')
    #print "py2exe extension module", __name__, "->", path
    mod = imp.load_dynamic(__name__, path)
##    mod.frozen = 1
__load()
del __load



#  Copyright (c) 2009-2010, Cloud Matrix Pty. Ltd.
#  All rights reserved; available under the terms of the BSD License.
"""

  esky.bdist_esky.f_cxfreeze:  bdist_esky support for cx_Freeze

"""


import os
import re
import sys
import imp
import time
import zipfile
import marshal
import struct
import shutil
import inspect
import zipfile
import distutils

if sys.platform == "win32":
    from esky import winres


import cx_Freeze
import cx_Freeze.hooks
INITNAME = "cx_Freeze__init__"

import esky
from esky.util import is_core_dependency


def freeze(dist):
    """Freeze the given distribution data using cx_Freeze."""
    includes = dist.includes
    excludes = dist.excludes
    options = dist.freezer_options
    #  Merge in any encludes/excludes given in freezer_options
    for inc in options.pop("includes",()):
        includes.append(inc)
    for exc in options.pop("excludes",()):
        excludes.append(exc)
    if "esky" not in includes and "esky" not in excludes:
        includes.append("esky")
    if "pypy" not in includes and "pypy" not in excludes:
        excludes.append("pypy")
    #  cx_Freeze doesn't seem to respect __path__ properly; hack it so
    #  that the required distutils modules are always found correctly.
    def load_distutils(finder,module):
        module.path = distutils.__path__ + module.path
        finder.IncludeModule("distutils.dist")
    cx_Freeze.hooks.load_distutils = load_distutils
    #  Build kwds arguments out of the given freezer opts.
    kwds = {}
    for (nm,val) in options.iteritems():
        kwds[_normalise_opt_name(nm)] = val
    kwds["includes"] = includes
    kwds["excludes"] = excludes
    kwds["targetDir"] = dist.freeze_dir
    #  Build an Executable object for each script.
    #  To include the esky startup code, we write each to a tempdir.
    executables = []
    for exe in dist.get_executables():
        base = None
        if exe.gui_only and sys.platform == "win32":
            base = "Win32GUI"
        executables.append(cx_Freeze.Executable(exe.script,base=base,targetName=exe.name,icon=exe.icon,**exe._kwds))
    #  Freeze up the executables
    f = cx_Freeze.Freezer(executables,**kwds)
    f.Freeze()
    #  Copy data files into the freeze dir
    for (src,dst) in dist.get_data_files():
        dst = os.path.join(dist.freeze_dir,dst)
        dstdir = os.path.dirname(dst)
        if not os.path.isdir(dstdir):
            dist.mkpath(dstdir)
        dist.copy_file(src,dst)
    #  Copy package data into the library.zip
    #  For now, this only works if there's a shared "library.zip" file.
    if f.createLibraryZip:
        lib = zipfile.ZipFile(os.path.join(dist.freeze_dir,"library.zip"),"a")
        for (src,arcnm) in dist.get_package_data():
            lib.write(src,arcnm)
        lib.close()
    else:
        for (src,arcnm) in dist.get_package_data():
            err = "use of package_data currently requires createLibraryZip=True"
            raise RuntimeError(err)
    #  Create the bootstrap code, using custom code if specified.
    code_source = ["__name__ = '__main__'"]
    esky_name = re.escape(dist.distribution.get_name())
    code_source.append("__esky_name__ = '%s'" % (esky_name,))
    code_source.append(inspect.getsource(esky.bootstrap))
    if dist.compile_bootstrap_exes:
        if sys.platform == "win32":
            #  Unfortunately this doesn't work, because the cxfreeze exe
            #  contains frozen modules that are inaccessible to a bootstrapped
            #  interpreter.  Disabled until I figure out a workaround. :-(
            pass
            #  The pypy-compiled bootstrap exe will try to load a python env
            #  into its own process and run this "take2" code to bootstrap.
            #take2_code = code_source[1:]
            #take2_code.append(_CUSTOM_WIN32_CHAINLOADER)
            #take2_code.append(dist.get_bootstrap_code())
            #take2_code = compile("\n".join(take2_code),"<string>","exec")
            #take2_code = marshal.dumps(take2_code)
            #clscript = "import marshal; "
            #clscript += "exec marshal.loads(%r); " % (take2_code,)
            #clscript = clscript.replace("%","%%")
            #clscript += "chainload(\"%s\")"
            #  Here's the actual source for the compiled bootstrap exe.
            #from esky.bdist_esky import pypy_libpython
            #code_source.append(inspect.getsource(pypy_libpython))
            #code_source.append("_PYPY_CHAINLOADER_SCRIPT = %r" % (clscript,))
            #code_source.append(_CUSTOM_PYPY_CHAINLOADER)
        code_source.append(dist.get_bootstrap_code())
        code_source = "\n".join(code_source)
        for exe in dist.get_executables(normalise=False):
            if not exe.include_in_bootstrap_env:
                continue
            bsexe = dist.compile_to_bootstrap_exe(exe,code_source)
            if sys.platform == "win32":
                fexe = os.path.join(dist.freeze_dir,exe.name)
                winres.copy_safe_resources(fexe,bsexe)
    else:
        if sys.platform == "win32":
            code_source.append(_CUSTOM_WIN32_CHAINLOADER)
        code_source.append(dist.get_bootstrap_code())
        code_source.append("bootstrap()")
        code_source = "\n".join(code_source)
        maincode = imp.get_magic() + struct.pack("<i",0)
        maincode += marshal.dumps(compile(code_source,INITNAME+".py","exec"))
        #  Create code for a fake esky.bootstrap module
        eskycode = imp.get_magic() + struct.pack("<i",0)
        eskycode += marshal.dumps(compile("","esky/__init__.py","exec"))
        eskybscode = imp.get_magic() + struct.pack("<i",0)
        eskybscode += marshal.dumps(compile("","esky/bootstrap.py","exec"))
        #  Copy any core dependencies
        if "fcntl" not in sys.builtin_module_names:
            for nm in os.listdir(dist.freeze_dir):
                if nm.startswith("fcntl"):
                    dist.copy_to_bootstrap_env(nm)
        for nm in os.listdir(dist.freeze_dir):
            if is_core_dependency(nm):
                dist.copy_to_bootstrap_env(nm)
        #  Copy the loader program for each script into the bootstrap env, and
        #  append the bootstrapping code to it as a zipfile.
        for exe in dist.get_executables(normalise=False):
            if not exe.include_in_bootstrap_env:
                continue
            exepath = dist.copy_to_bootstrap_env(exe.name)
            bslib = zipfile.PyZipFile(exepath,"a",zipfile.ZIP_STORED)
            cdate = (2000,1,1,0,0,0)
            bslib.writestr(zipfile.ZipInfo(INITNAME+".pyc",cdate),maincode)
            bslib.writestr(zipfile.ZipInfo("esky/__init__.pyc",cdate),eskycode)
            bslib.writestr(zipfile.ZipInfo("esky/bootstrap.pyc",cdate),eskybscode)
            bslib.close()


def _normalise_opt_name(nm):
    """Normalise option names for cx_Freeze.

    This allows people to specify options named like "opt-name" and have
    them converted to the "optName" format used internally by cx_Freeze.
    """
    bits = nm.split("-")
    for i in xrange(1,len(bits)):
        if bits[i]:
            bits[i] = bits[i][0].upper() + bits[i][1:]
    return "".join(bits)


#  On Windows, execv is flaky and expensive.  If the chainloader is the same
#  python version as the target exe, we can munge sys.path to bootstrap it
#  into the existing process.
if sys.version_info[0] < 3:
    EXEC_STATEMENT = "exec code in globals()"
else:
    EXEC_STATEMENT = "exec(code,globals())"

_CUSTOM_WIN32_CHAINLOADER = """

_orig_chainload = _chainload
def _chainload(target_dir):
  mydir = dirname(sys.executable)
  pydll = "python%%s%%s.dll" %% sys.version_info[:2]
  if not exists(pathjoin(target_dir,pydll)):
      _orig_chainload(target_dir)
  else:
      sys.bootstrap_executable = sys.executable
      sys.executable = pathjoin(target_dir,basename(sys.executable))
      verify(sys.executable)
      sys.prefix = sys.prefix.replace(mydir,target_dir)
      sys.argv[0] = sys.executable
      for i in range(len(sys.path)):
          sys.path[i] = sys.path[i].replace(mydir,target_dir)
      curdir = getcwd()
      newdir = curdir.replace(mydir,target_dir)
      if newdir != curdir:
          nt.chdir(newdir)
      import zipimport
      for init_path in sys.path:
          verify(init_path)
          try:
              importer = zipimport.zipimporter(init_path)
              code = importer.get_code("%s")
          except ImportError:
              pass
          else:
              sys.modules.pop("esky",None)
              sys.modules.pop("esky.bootstrap",None)
              #  Adjust various cxfreeze global vars
              global DIR_NAME, FILE_NAME, EXCLUSIVE_ZIP_FILE_NAME
              global SHARED_ZIP_FILE_NAME, INITSCRIPT_ZIP_FILE_NAME
              # TODO: are these derivations correct?
              FILE_NAME = sys.executable
              DIR_NAME = dirname(sys.executable)
              if FILE_NAME.endswith(".exe"):
                  EXCLUSIVE_ZIP_FILE_NAME = pathjoin(DIR_NAME,basename(FILE_NAME)[:-4]+".zip")
              else:
                  EXCLUSIVE_ZIP_FILE_NAME = pathjoin(DIR_NAME,basename(FILE_NAME)+".zip")
              SHARED_ZIP_FILE_NAME = pathjoin(DIR_NAME,"library.zip")
              INITSCRIPT_ZIP_FILE_NAME = init_path
              try:
                  %s
              except zipimport.ZipImportError:
                  e = sys.exc_info()[1]
                  #  If it can't find the __main__{sys.executable} script,
                  #  the user might be running from a backup exe file.
                  #  Fall back to original chainloader to attempt workaround.
                  if e.message.endswith("__main__'"):
                      _orig_chainload(target_dir)
                  raise
              sys.exit(0)
      else:
          _orig_chainload(target_dir)
""" % (INITNAME,EXEC_STATEMENT)


#  On Windows, execv is flaky and expensive.  Since the pypy-compiled bootstrap
#  exe doesn't have a python runtime, it needs to chainload the one from the
#  target version dir before trying to bootstrap in-process.
_CUSTOM_PYPY_CHAINLOADER = """

_orig_chainload = _chainload
def _chainload(target_dir):
  mydir = dirname(sys.executable)
  pydll = "python%s%s.dll" % sys.version_info[:2]
  if not exists(pathjoin(target_dir,pydll)):
      _orig_chainload(target_dir)
  else:
      py = libpython(pydll)

      #Py_NoSiteFlag = 1;
      #Py_FrozenFlag = 1;
      #Py_IgnoreEnvironmentFlag = 1;

      py.SetPythonHome("")
      py.Initialize()
      # TODO: can't get this through pypy's type annotator.
      # going to fudge it in python instead :-)
      #py.Sys_SetArgv(list(sys.argv))
      syspath = py.GetProgramFullPath() + ";"
      sysfilenm = basename(py.GetProgramFullPath())
      i = 0
      while i < len(sysfilenm) and sysfilenm[i:] != ".exe":
          i += 1
      sysfilenm = sysfilenm[:i]
      sysdir = dirname(py.GetProgramFullPath())
      syspath += sysdir + "\\%s.zip;" % (sysfilenm,)
      syspath += sysdir + "\\library.zip;"
      syspath += sysdir
      py.Sys_SetPath(syspath);
      #  Escape any double-quotes in sys.argv, so we can easily
      #  include it in a python-level string.
      new_argvs = []
      for arg in sys.argv:
          new_argvs.append('"' + arg.replace('"','\\"') + '"')
      new_argv = "[" + ",".join(new_argvs) + "]"
      py.Run_SimpleString("import sys; sys.argv = %s" % (new_argv,))
      py.Run_SimpleString("import sys; sys.frozen = 'cxfreeze'" % (new_argv,))
      globals = py.Dict_New()
      py.Dict_SetItemString(globals,"__builtins__",py.Eval_GetBuiltins())
      esc_target_dir_chars = []
      for c in target_dir:
          if c == "\\\\":
              esc_target_dir_chars.append("\\\\")
          esc_target_dir_chars.append(c)
      esc_target_dir = "".join(esc_target_dir_chars)
      script = _PYPY_CHAINLOADER_SCRIPT % (esc_target_dir,)
      py.Run_String(script,py.file_input,globals)
      py.Finalize()
      sys.exit(0)

"""





"""

  esky.bdist_esky.pypy_winres:  access win32 exe resources in rpython


This module provides some functions for accessing win32 exe resources from
rpython code.  It's a trimmed-down version of the esky.winres module with
just enough functionality to get the py2exe compiled bootstrapper working.

"""

from pypy.rlib import clibffi
from pypy.rpython.lltypesystem import rffi, lltype
from pypy.rlib import rwin32


LOAD_LIBRARY_AS_DATAFILE = 0x00000002


k32_LoadLibraryExA = rwin32.winexternal("LoadLibraryExA",[rffi.CCHARP,rwin32.HANDLE,rwin32.DWORD],rwin32.HANDLE)
k32_FindResourceExA = rwin32.winexternal("FindResourceExA",[rwin32.HANDLE,rffi.CCHARP,rwin32.DWORD,rwin32.DWORD],rwin32.HANDLE)
k32_SizeofResource = rwin32.winexternal("SizeofResource",[rwin32.HANDLE,rwin32.HANDLE],rwin32.DWORD)
k32_LoadResource = rwin32.winexternal("LoadResource",[rwin32.HANDLE,rwin32.HANDLE],rwin32.HANDLE)
k32_LockResource = rwin32.winexternal("LockResource",[rwin32.HANDLE],rffi.CCHARP)
k32_FreeLibrary = rwin32.winexternal("FreeLibrary",[rwin32.HANDLE],rwin32.BOOL)


def load_resource(filename,resname,resid,reslang):
    """Load the named resource from the given file.

    The filename and resource name must be ascii strings, and the resid and
    reslang must be integers.
    """
    l_handle = k32_LoadLibraryExA(filename,rffi.cast(rwin32.HANDLE,0),LOAD_LIBRARY_AS_DATAFILE)
    if not l_handle:
        raise WindowsError(rwin32.GetLastError(),"LoadLibraryExW failed")
    try:
        r_handle = k32_FindResourceExA(l_handle,resname,resid,reslang)
        if not r_handle:
            raise WindowsError(rwin32.GetLastError(),"FindResourceExA failed")
        r_size = k32_SizeofResource(l_handle,r_handle)
        if not r_size:
            raise WindowsError(rwin32.GetLastError(),"SizeofResource failed")
        r_info = k32_LoadResource(l_handle,r_handle)
        if not r_info:
            raise WindowsError(rwin32.GetLastError(),"LoadResource failed")
        r_ptr = k32_LockResource(r_info)
        if not r_ptr:
            raise WindowsError(rwin32.GetLastError(),"LockResource failed")
        return rffi.charpsize2str(r_ptr,r_size)
    finally:
        if not k32_FreeLibrary(l_handle):
            raise WindowsError(rwin32.GetLastError(),"FreeLibrary failed")


def load_resource_pystr(py,filename,resname,resid,reslang):
    """Load the named resource from the given file as a python-level string

    The filename and resource name must be ascii strings, and the resid and
    reslang must be integers.

    This uses the given python dll object to load the data directly into 
    a python string, saving a lot of copying and carrying on.
    """
    l_handle = k32_LoadLibraryExA(filename,rffi.cast(rwin32.HANDLE,0),LOAD_LIBRARY_AS_DATAFILE)
    if not l_handle:
        raise WindowsError(rwin32.GetLastError(),"LoadLibraryExW failed")
    try:
        r_handle = k32_FindResourceExA(l_handle,resname,resid,reslang)
        if not r_handle:
            raise WindowsError(rwin32.GetLastError(),"FindResourceExA failed")
        r_size = k32_SizeofResource(l_handle,r_handle)
        if not r_size:
            raise WindowsError(rwin32.GetLastError(),"SizeofResource failed")
        r_info = k32_LoadResource(l_handle,r_handle)
        if not r_info:
            raise WindowsError(rwin32.GetLastError(),"LoadResource failed")
        r_ptr = k32_LockResource(r_info)
        if not r_ptr:
            raise WindowsError(rwin32.GetLastError(),"LockResource failed")
        s = py.String_FromStringAndSize(None,r_size)
        buf = py.String_AsString(s)
        memcpy(buf,rffi.cast(rffi.VOIDP,r_ptr),r_size)
        return s
    finally:
        if not k32_FreeLibrary(l_handle):
            raise WindowsError(rwin32.GetLastError(),"FreeLibrary failed")


def memcpy(target,source,n):
    impl = clibffi.CDLL(clibffi.get_libc_name()).getpointer("memcpy",[clibffi.ffi_type_pointer,clibffi.ffi_type_pointer,clibffi.ffi_type_uint],clibffi.ffi_type_void)
    impl.push_arg(target)
    impl.push_arg(source)
    impl.push_arg(n)
    impl.call(lltype.Void)
   




#  Copyright (c) 2009-2010, Cloud Matrix Pty. Ltd.
#  All rights reserved; available under the terms of the BSD License.
"""

  esky.bdist_esky.f_bbfreeze:  bdist_esky support for bbfreeze

"""


import os
import re
import sys
import imp
import time
import zipfile
import tempfile
import marshal
import struct
import shutil
import inspect
import zipfile

if sys.platform == "win32":
    from esky import winres


import bbfreeze

import esky
from esky.util import is_core_dependency


def freeze(dist):
    """Freeze the given distribution data using bbfreeze."""
    includes = dist.includes
    excludes = dist.excludes
    options = dist.freezer_options
    #  Merge in any encludes/excludes given in freezer_options
    for inc in options.pop("includes",()):
        includes.append(inc)
    for exc in options.pop("excludes",()):
        excludes.append(exc)
    if "pypy" not in includes and "pypy" not in excludes:
        excludes.append("pypy")
    #  Freeze up the given scripts
    f = bbfreeze.Freezer(dist.freeze_dir,includes=includes,excludes=excludes)
    for (nm,val) in options.iteritems():
        setattr(f,nm,val)
    f.addModule("esky")
    tdir = tempfile.mkdtemp()
    try:
        for exe in dist.get_executables():
            f.addScript(exe.script,gui_only=exe.gui_only)
        if "include_py" not in options:
            f.include_py = False
        if "linkmethod" not in options:
            #  Since we're going to zip it up, the benefits of hard-
            #  or sym-linking the loader exe will mostly be lost.
            f.linkmethod = "loader"
        f()
    finally:
        shutil.rmtree(tdir)
    #  Copy data files into the freeze dir
    for (src,dst) in dist.get_data_files():
        dst = os.path.join(dist.freeze_dir,dst)
        dstdir = os.path.dirname(dst)
        if not os.path.isdir(dstdir):
            dist.mkpath(dstdir)
        dist.copy_file(src,dst)
    #  Copy package data into the library.zip
    lib = zipfile.ZipFile(os.path.join(dist.freeze_dir,"library.zip"),"a")
    for (src,arcnm) in dist.get_package_data():
        lib.write(src,arcnm)
    lib.close()
    #  Create the bootstrap code, using custom code if specified.
    #  For win32 we include a special chainloader that can suck the selected
    #  version into the running process rather than spawn a new proc.
    code_source = ["__name__ = '__main__'"]
    esky_name = re.escape(dist.distribution.get_name())
    code_source.append("__esky_name__ = '%s'" % (esky_name,))
    code_source.append(inspect.getsource(esky.bootstrap))
    if dist.compile_bootstrap_exes:
        if sys.platform == "win32":
            #  The pypy-compiled bootstrap exe will try to load a python env
            #  into its own process and run this "take2" code to bootstrap.
            take2_code = code_source[1:]
            take2_code.append(_CUSTOM_WIN32_CHAINLOADER)
            take2_code.append(dist.get_bootstrap_code())
            take2_code = compile("\n".join(take2_code),"<string>","exec")
            take2_code = marshal.dumps(take2_code)
            clscript = "import marshal; "
            clscript += "exec marshal.loads(%r); " % (take2_code,)
            clscript = clscript.replace("%","%%")
            clscript += "chainload(\"%s\")"
            #  Here's the actual source for the compiled bootstrap exe.
            from esky.bdist_esky import pypy_libpython
            code_source.append(inspect.getsource(pypy_libpython))
            code_source.append("_PYPY_CHAINLOADER_SCRIPT = %r" % (clscript,))
            code_source.append(_CUSTOM_PYPY_CHAINLOADER)
        code_source.append(dist.get_bootstrap_code())
        code_source = "\n".join(code_source)
        for exe in dist.get_executables(normalise=False):
            if not exe.include_in_bootstrap_env:
                continue
            bsexe = dist.compile_to_bootstrap_exe(exe,code_source)
            if sys.platform == "win32":
                fexe = os.path.join(dist.freeze_dir,exe.name)
                winres.copy_safe_resources(fexe,bsexe)
        #  We may also need the bundled MSVCRT libs
        if sys.platform == "win32":
            for nm in os.listdir(dist.freeze_dir):
                if is_core_dependency(nm) and nm.startswith("Microsoft"):
                    dist.copy_to_bootstrap_env(nm)
    else:
        if sys.platform == "win32":
            code_source.append(_CUSTOM_WIN32_CHAINLOADER)
        code_source.append(dist.get_bootstrap_code())
        code_source.append("bootstrap()")
        code_source = "\n".join(code_source)
        #  For non-compiled bootstrap exe, store the bootstrapping code
        #  into the library.zip as __main__.
        maincode = imp.get_magic() + struct.pack("<i",0)
        maincode += marshal.dumps(compile(code_source,"__main__.py","exec"))
        #  Create code for a fake esky.bootstrap module
        eskycode = imp.get_magic() + struct.pack("<i",0)
        eskycode += marshal.dumps(compile("","esky/__init__.py","exec"))
        eskybscode = imp.get_magic() + struct.pack("<i",0)
        eskybscode += marshal.dumps(compile("","esky/bootstrap.py","exec"))
        #  Store bootstrap code as __main__ in the bootstrap library.zip.
        #  The frozen library.zip might have the loader prepended to it, but
        #  that gets overwritten here.
        bslib_path = dist.copy_to_bootstrap_env("library.zip")
        bslib = zipfile.PyZipFile(bslib_path,"w",zipfile.ZIP_STORED)
        cdate = (2000,1,1,0,0,0)
        bslib.writestr(zipfile.ZipInfo("__main__.pyc",cdate),maincode)
        bslib.writestr(zipfile.ZipInfo("esky/__init__.pyc",cdate),eskycode)
        bslib.writestr(zipfile.ZipInfo("esky/bootstrap.pyc",cdate),eskybscode)
        bslib.close()
        #  Copy any core dependencies
        if "fcntl" not in sys.builtin_module_names:
            for nm in os.listdir(dist.freeze_dir):
                if nm.startswith("fcntl"):
                    dist.copy_to_bootstrap_env(nm)
        for nm in os.listdir(dist.freeze_dir):
            if is_core_dependency(nm):
                dist.copy_to_bootstrap_env(nm)
        #  Copy the bbfreeze interpreter if necessary
        if f.include_py:
            if sys.platform == "win32":
                dist.copy_to_bootstrap_env("py.exe")
            else:
                dist.copy_to_bootstrap_env("py")
        #  Copy the loader program for each script.
        #  We explicitly strip the loader binaries, in case they were made
        #  by linking to the library.zip.
        for exe in dist.get_executables(normalise=False):
            if not exe.include_in_bootstrap_env:
                continue
            exepath = dist.copy_to_bootstrap_env(exe.name)
            f.stripBinary(exepath)


#  On Windows, execv is flaky and expensive.  If the chainloader is the same
#  python version as the target exe, we can munge sys.path to bootstrap it
#  into the existing process.
_CUSTOM_WIN32_CHAINLOADER = """
_orig_chainload = _chainload
def _chainload(target_dir):
  mydir = dirname(sys.executable)
  pydll = "python%s%s.dll" % sys.version_info[:2]
  if not exists(pathjoin(target_dir,pydll)):
      _orig_chainload(target_dir)
  else:
      sys.bootstrap_executable = sys.executable
      sys.executable = pathjoin(target_dir,basename(sys.executable))
      verify(sys.executable)
      sys.prefix = sys.prefix.replace(mydir,target_dir)
      sys.argv[0] = sys.executable
      for i in xrange(len(sys.path)):
          sys.path[i] = sys.path[i].replace(mydir,target_dir)
      curdir = getcwd()
      newdir = curdir.replace(mydir,target_dir)
      if newdir != curdir:
          nt.chdir(newdir)
      try:
          verify(sys.path[0])
          import zipimport
          importer = zipimport.zipimporter(sys.path[0])
          code = importer.get_code("__main__")
      except ImportError:
          _orig_chainload(target_dir)
      else:
          sys.modules.pop("esky",None)
          sys.modules.pop("esky.bootstrap",None)
          try:
              exec code in {"__name__":"__main__"}
          except zipimport.ZipImportError, e:
              #  If it can't find the __main__{sys.executable} script,
              #  the user might be running from a backup exe file.
              #  Fall back to original chainloader to attempt workaround.
              if e.message.startswith("can't find module '__main__"):
                  _orig_chainload(target_dir)
              raise
          sys.exit(0)
"""

#  On Windows, execv is flaky and expensive.  Since the pypy-compiled bootstrap
#  exe doesn't have a python runtime, it needs to chainload the one from the
#  target version dir before trying to bootstrap in-process.
_CUSTOM_PYPY_CHAINLOADER = """

_orig_chainload = _chainload
def _chainload(target_dir):
  mydir = dirname(sys.executable)
  pydll = pathjoin(target_dir,"python%s%s.dll" % sys.version_info[:2])
  if not exists(pydll):
      _orig_chainload(target_dir)
  else:
      py = libpython(pydll)


      py.Set_NoSiteFlag(1)
      py.Set_FrozenFlag(1)
      py.Set_IgnoreEnvironmentFlag(1)

      py.SetPythonHome("")
      py.Initialize()
      # TODO: can't get this through pypy's type annotator.
      # going to fudge it in python instead :-)
      #py.Sys_SetArgv(list(sys.argv))
      syspath = dirname(py.GetProgramFullPath());
      syspath = syspath + "\\library.zip;" + syspath
      py.Sys_SetPath(syspath);
      #  Escape any double-quotes in sys.argv, so we can easily
      #  include it in a python-level string.
      new_argvs = []
      for arg in sys.argv:
          new_argvs.append('"' + arg.replace('"','\\"') + '"')
      new_argv = "[" + ",".join(new_argvs) + "]"
      py.Run_SimpleString("import sys; sys.argv = %s" % (new_argv,))
      py.Run_SimpleString("import sys; sys.frozen = 'bbfreeze'" % (new_argv,))
      globals = py.Dict_New()
      py.Dict_SetItemString(globals,"__builtins__",py.Eval_GetBuiltins())
      esc_target_dir_chars = []
      for c in target_dir:
          if c == "\\\\":
              esc_target_dir_chars.append("\\\\")
          esc_target_dir_chars.append(c)
      esc_target_dir = "".join(esc_target_dir_chars)
      script = _PYPY_CHAINLOADER_SCRIPT % (esc_target_dir,)
      py.Run_String(script,py.file_input,globals)
      py.Finalize()
      sys.exit(0)

"""



#  Copyright (c) 2009-2010, Cloud Matrix Pty. Ltd.
#  All rights reserved; available under the terms of the BSD License.
"""

  esky.fstransact: best-effort support for transactional filesystem operations

This module provides a uniform interface to various platform-specific 
mechanisms for doing transactional filesystem operations.  On platforms where
transactions are not supported, it falls back to doing things one operation
at a time.

Currently supported platforms are:

    * Windows Vista and later, using MoveFileTransacted and friends
    * err..that's it for the moment, actually

Although transactions are not supported on POSIX platforms, the way Esky
structures its filesystem operations means that the program is always safe
as long as you can atomically replace a file.

"""

import os
import sys
import shutil

from esky.util import lazy_import


@lazy_import
def _fallback():
    import esky.fstransact.fallback
    return esky.fstransact.fallback

@lazy_import
def _win32txf():
    try:
        import esky.fstransact.win32txf
    except ImportError:
        return None
    else:
        return esky.fstransact.win32txf


def FSTransaction(root=None):
    """Factory function returning FSTransaction objects.

    This factory function takes the root path within which file operations
    will be performed, and returns an appropriate FSTransaction object that
    provides best-effort transactional operations for that root.
    """
    #  Try to use TxF on win32.  This might fail because it's not available,
    #  or because the target filesystem doesn't support it.
    if sys.platform == "win32" and _win32txf:
        try:
            return _win32txf.FSTransaction(root)
        except WindowsError, e:
            if e.winerror != _win32txf.ERROR_TRANSACTIONAL_OPEN_NOT_ALLOWED:
                raise
    #  If all else fails, use the fallback implementation.
    return _fallback.FSTransaction(root)





#  Copyright (c) 2009-2010, Cloud Matrix Pty. Ltd.
#  All rights reserved; available under the terms of the BSD License.
"""

  esky.bdist_esky:  distutils command to freeze apps in esky format

Importing this module makes "bdist_esky" available as a distutils command.
This command will freeze the given scripts and package them into a zipfile
named with the application name, version and platform.

The resulting zipfile is conveniently in the format expected by the class
DefaultVersionFinder.  It will be named "appname-version.platform.zip"

"""

from __future__ import with_statement


import os
import re
import sys
import shutil
import zipfile
import tempfile
import hashlib
import inspect
from glob import glob

import distutils.command
from distutils.core import Command
from distutils.util import convert_path

import esky.patch
from esky.util import get_platform, is_core_dependency, create_zipfile, \
                      split_app_version, join_app_version, ESKY_CONTROL_DIR, \
                      ESKY_APPDATA_DIR, really_rmtree

if sys.platform == "win32":
    from esky import winres
    from xml.dom import minidom

try:
    from esky.bdist_esky import pypyc
except ImportError, e:
    pypyc = None
    PYPYC_ERROR = e
    COMPILED_BOOTSTRAP_CACHE = None
else:
    COMPILED_BOOTSTRAP_CACHE = os.path.dirname(__file__)
    if not os.path.isdir(COMPILED_BOOTSTRAP_CACHE):
        COMPILED_BOOTSTRAP_CACHE = None



#  setuptools likes to be imported before anything else that
#  might monkey-patch distutils.  We don't actually use it,
#  this is just to avoid errors with cx_Freeze.
try:
    import setuptools
except ImportError:
    pass


_FREEZERS = {}
try:
    from esky.bdist_esky import f_py2exe
    _FREEZERS["py2exe"] = f_py2exe
except ImportError:
    _FREEZERS["py2exe"] = None
try:
    from esky.bdist_esky import f_py2app
    _FREEZERS["py2app"] = f_py2app
except ImportError:
    _FREEZERS["py2app"] = None
try:
    from esky.bdist_esky import f_bbfreeze
    _FREEZERS["bbfreeze"] = f_bbfreeze
except ImportError:
    _FREEZERS["bbfreeze"] = None
try:
    from esky.bdist_esky import f_cxfreeze
    _FREEZERS["cxfreeze"] = f_cxfreeze
    _FREEZERS["cx_Freeze"] = f_cxfreeze
    _FREEZERS["cx_freeze"] = f_cxfreeze
except ImportError:
    _FREEZERS["cxfreeze"] = None
    _FREEZERS["cx_Freeze"] = None
    _FREEZERS["cx_freeze"] = None


class Executable(unicode):
    """Class to hold information about a specific executable.

    This class provides a uniform way to specify extra meta-data about
    a frozen executable.  By setting various keyword arguments, you can
    specify e.g. the icon, and whether it is a gui-only script.

    Some freezer modules require all items in the "scripts" argument to
    be strings naming real files.  This is therefore a subclass of unicode,
    and if it refers only to in-memory code then its string value will be
    the path to this very file.  I know it's ugly, but it works.
    """

    def __new__(cls,script,**kwds):
        if isinstance(script,basestring):
            return unicode.__new__(cls,script)
        else:
            return unicode.__new__(cls,__file__)

    def __init__(self,script,name=None,icon=None,gui_only=None,
                      include_in_bootstrap_env=True,**kwds):
        unicode.__init__(self)
        if isinstance(script,Executable):
            script = script.script
            if name is None:
                name = script.name
            if gui_only is None:
                gui_only = script.gui_only
        if not isinstance(script,basestring):
            if name is None:
                raise TypeError("Must specify name if script is not a file")
        self.script = script
        self.include_in_bootstrap_env = include_in_bootstrap_env
        self.icon = icon
        self._name = name
        self._gui_only = gui_only
        self._kwds = kwds

    @property
    def name(self):
        if self._name is not None:
            nm = self._name
        else:
            if not isinstance(self.script,basestring):
                raise TypeError("Must specify name if script is not a file")
            nm = os.path.basename(self.script)
            if nm.endswith(".py"):
                nm = nm[:-3]
            elif nm.endswith(".pyw"):
                nm = nm[:-4]
        if sys.platform == "win32" and not nm.endswith(".exe"):
            nm += ".exe"
        return nm

    @property
    def gui_only(self):
        if self._gui_only is None:
            if not isinstance(self.script,basestring):
                return False
            else:
                return self.script.endswith(".pyw")
        else:
            return self._gui_only



class bdist_esky(Command):
    """Create a frozen application in 'esky' format.

    This distutils command can be used to freeze an application in the
    format expected by esky.  It interprets the following standard 
    distutils options:

       scripts:  list of scripts to freeze as executables;
                 to make a gui-only script, name it 'script.pyw'

       data_files:  copied into the frozen app directory

       package_data:  copied into library.zip alongside the module code

    To further customize the behaviour of the bdist_esky command, you can
    specify the following custom options:

        includes:  a list of modules to explicitly include in the freeze

        excludes:  a list of modules to explicitly exclude from the freeze

        freezer_module:  name of freezer module to use; currently py2exe,
                         py2app,  bbfreeze and cx-freeze are supported.

        freezer_options: dict of options to pass through to the underlying
                         freezer module.

        bootstrap_module:  a custom module to use for esky bootstrapping;
                           the default calls esky.bootstrap.bootstrap()

        bootstrap_code:  a custom code string to use for esky bootstrapping;
                         this precludes the use of the bootstrap_module option.
                         If a non-string object is given, its source is taken
                         using inspect.getsource().

        compile_bootstrap_exes:  whether to compile the bootstrapping code to a
                                 stand-alone exe; this requires PyPy installed
                                 and the bootstrap code to be valid RPython.
                                 When false, the bootstrap env will use a
                                 trimmed-down copy of the freezer module exe.

        dont_run_startup_hooks:  don't force all executables to call
                                 esky.run_startup_hooks() on startup.

        bundle_msvcrt:  whether to bundle the MSVCRT DLLs, manifest files etc
                        as a private assembly.  The default is False; only
                        those with a valid license to redistriute these files
                        should enable it.



        pre_freeze_callback:  function to call just before starting to freeze
                              the application; this is a good opportunity to
                              customize the bdist_esky instance.

        pre_zip_callback:  function to call just before starting to zip up
                           the frozen application; this is a good opportunity
                           to e.g. sign the resulting executables.
    
    """

    description = "create a frozen app in 'esky' format"

    user_options = [
                    ('dist-dir=', 'd',
                     "directory to put final built distributions in"),
                    ('freezer-module=', None,
                     "module to use for freezing the application"),
                    ('freezer-options=', None,
                     "options to pass to the underlying freezer module"),
                    ('bootstrap-module=', None,
                     "module to use for bootstrapping the application"),
                    ('bootstrap-code=', None,
                     "code to use for bootstrapping the application"),
                    ('compile-bootstrap-exes=', None,
                     "whether to compile the bootstrapping exes with pypy"),
                    ('bundle-msvcrt=', None,
                     "whether to bundle MSVCRT as private assembly"),
                    ('includes=', None,
                     "list of modules to specifically include"),
                    ('excludes=', None,
                     "list of modules to specifically exclude"),
                    ('dont-run-startup-hooks=', None,
                     "don't force execution of esky.run_startup_hooks()"),
                    ('pre-freeze-callback=', None,
                     "function to call just before starting to freeze the app"),
                    ('pre-zip-callback=', None,
                     "function to call just before starting to zip up the app"),
                    ('enable-appdata-dir=', None,
                     "enable new 'appdata' directory layout (will go away after the 0.9.X series)"),
                   ]

    boolean_options = ["bundle-msvcrt","dont-run-startup-hooks","compile-bootstrap-exes","enable-appdata-dir"]

    def initialize_options(self):
        self.dist_dir = None
        self.includes = []
        self.excludes = []
        self.freezer_module = None
        self.freezer_options = {}
        self.bundle_msvcrt = False
        self.dont_run_startup_hooks = False
        self.bootstrap_module = None
        self.bootstrap_code = None
        self.compile_bootstrap_exes = False
        self._compiled_exes = {}
        self.pre_freeze_callback = None
        self.pre_zip_callback = None
        self.enable_appdata_dir = False

    def finalize_options(self):
        self.set_undefined_options('bdist',('dist_dir', 'dist_dir'))
        if self.compile_bootstrap_exes and pypyc is None:
            raise PYPYC_ERROR
        if self.freezer_module is None:
            for freezer_module in ("py2exe","py2app","bbfreeze","cxfreeze"):
                self.freezer_module = _FREEZERS[freezer_module]
                if self.freezer_module is not None:
                    break
            else:
                err = "no supported freezer modules found"
                err += " (try installing bbfreeze)"
                raise RuntimeError(err)
        else:
            try:
                freezer = _FREEZERS[self.freezer_module]
            except KeyError:
                err = "freezer module not supported: '%s'"
                err = err % (self.freezer_module,)
                raise RuntimeError(err)
            else:
                if freezer is None:
                    err = "freezer module not found: '%s'"
                    err = err % (self.freezer_module,)
                    raise RuntimeError(err)
            self.freezer_module = freezer
        if isinstance(self.pre_freeze_callback,basestring):
            self.pre_freeze_callback = self._name2func(self.pre_freeze_callback)
        if isinstance(self.pre_zip_callback,basestring):
            self.pre_zip_callback = self._name2func(self.pre_zip_callback)

    def _name2func(self,name):
        """Convert a dotted name into a function reference."""
        if "." not in name:
            return globals()[name]
        modname,funcname = name.rsplit(".",1)
        mod = __import__(modname,fromlist=[funcname])
        return getattr(mod,funcname)

    def run(self):
        self.tempdir = tempfile.mkdtemp()
        try:
            self._run()
        finally:
            really_rmtree(self.tempdir)

    def _run(self):
        self._run_initialise_dirs()
        if self.pre_freeze_callback is not None:
            self.pre_freeze_callback(self)
        self._run_freeze_scripts()
        if self.pre_zip_callback is not None:
            self.pre_zip_callback(self)
        self._run_create_zipfile()

    def _run_initialise_dirs(self):
        """Create the dirs into which to freeze the app."""
        fullname = self.distribution.get_fullname()
        platform = get_platform()
        self.bootstrap_dir = os.path.join(self.dist_dir,
                                          "%s.%s"%(fullname,platform,))
        if self.enable_appdata_dir:
            self.freeze_dir = os.path.join(self.bootstrap_dir,ESKY_APPDATA_DIR,
                                           "%s.%s"%(fullname,platform,))
        else:
            self.freeze_dir = os.path.join(self.bootstrap_dir,
                                           "%s.%s"%(fullname,platform,))
        if os.path.exists(self.bootstrap_dir):
            really_rmtree(self.bootstrap_dir)
        os.makedirs(self.freeze_dir)

    def _run_freeze_scripts(self):
        """Call the selected freezer module to freeze the scripts."""
        fullname = self.distribution.get_fullname()
        platform = get_platform()
        self.freezer_module.freeze(self)
        if platform != "win32":
            lockfile = os.path.join(self.freeze_dir,ESKY_CONTROL_DIR,"lockfile.txt")
            with open(lockfile,"w") as lf:
                lf.write("this file is used by esky to lock the version dir\n")

    def _run_create_zipfile(self):
        """Zip up the final distribution."""
        print "zipping up the esky"
        fullname = self.distribution.get_fullname()
        platform = get_platform()
        zfname = os.path.join(self.dist_dir,"%s.%s.zip"%(fullname,platform,))
        if hasattr(self.freezer_module,"zipit"):
            self.freezer_module.zipit(self,self.bootstrap_dir,zfname)
        else:
            create_zipfile(self.bootstrap_dir,zfname,compress=True)
        really_rmtree(self.bootstrap_dir)

    def _obj2code(self,obj):
        """Convert an object to some python source code.

        Iterables are flattened, None is elided, strings are included verbatim,
        open files are read and anything else is passed to inspect.getsource().
        """
        if obj is None:
            return ""
        if isinstance(obj,basestring):
            return obj
        if hasattr(obj,"read"):
            return obj.read()
        try:
            return "\n\n\n".join(self._obj2code(i) for i in obj)
        except TypeError:
            return inspect.getsource(obj)

    def get_bootstrap_code(self):
        """Get any extra code to be executed by the bootstrapping exe.

        This method interprets the bootstrap-code and bootstrap-module settings
        to construct any extra bootstrapping code that must be executed by
        the frozen bootstrap executable.  It is returned as a string.
        """
        bscode = self.bootstrap_code
        if bscode is None:
            if self.bootstrap_module is not None:
                bscode = __import__(self.bootstrap_module)
                for submod in self.bootstrap_module.split(".")[1:]:
                    bscode = getattr(bscode,submod)
        bscode = self._obj2code(bscode)
        return bscode

    def get_executables(self,normalise=True):
        """Yield a normalised Executable instance for each script to be frozen.

        If "normalise" is True (the default) then the user-provided scripts
        will be rewritten to decode any non-filename items specified as part
        of the script, and to include the esky startup code.  If the freezer
        has a better way of doing these things, it should pass normalise=False.
        """
        if normalise:
            if not os.path.exists(os.path.join(self.tempdir,"scripts")):
                os.mkdir(os.path.join(self.tempdir,"scripts"))
        if self.distribution.has_scripts():
            for s in self.distribution.scripts:
                if isinstance(s,Executable):
                    exe = s
                else:
                    exe = Executable(s)
                if normalise:
                    #  Give the normalised script file a name matching that
                    #  specified, since some freezers only take the filename.
                    name = exe.name
                    if sys.platform == "win32" and name.endswith(".exe"):
                        name = name[:-4]
                    if exe.endswith(".pyw"):
                        ext = ".pyw"
                    else:
                        ext = ".py"
                    script = os.path.join(self.tempdir,"scripts",name+ext)
                    #  Get the code for the target script.
                    #  If it's a single string then interpret it as a filename,
                    #  otherwise feed it into the _obj2code logic.
                    if isinstance(exe.script,basestring):
                        with open(exe.script,"rt") as f:
                            code = f.read()
                    else:
                        code = self._obj2code(exe.script)
                    #  Check that the code actually compiles - sometimes it
                    #  can be hard to get a good message out of the freezer.
                    compile(code,"","exec")
                    #  Augment the given code with special esky-related logic.
                    with open(script,"wt") as fOut:
                        lines = (ln+"\n" for ln in code.split("\n"))
                        #  Keep any leading comments and __future__ imports
                        #  at the start of the file.
                        for ln in lines:
                            if ln.strip():
                                if not ln.strip().startswith("#"):
                                    if "__future__" not in ln:
                                        break
                            fOut.write(ln)
                        #  Run the startup hooks before any actual code.
                        if not self.dont_run_startup_hooks:
                            fOut.write("import esky\n")
                            fOut.write("esky.run_startup_hooks()\n")
                            fOut.write("\n")
                        #  Then just include the rest of the script code.
                        fOut.write(ln)
                        for ln in lines:
                            fOut.write(ln)
                    new_exe = Executable(script)
                    new_exe.__dict__.update(exe.__dict__)
                    new_exe.script = script
                    exe = new_exe
                yield exe

    def get_data_files(self):
        """Yield (source,destination) tuples for data files.

        This method generates the names of all data file to be included in
        the frozen app.  They should be placed directly into the freeze
        directory as raw files.
        """
        fdir = self.freeze_dir
        if sys.platform == "win32" and self.bundle_msvcrt:
            for (src,dst) in self.get_msvcrt_private_assembly_files():
                yield (src,dst)
        if self.distribution.data_files:
            for datafile in self.distribution.data_files:
                #  Plain strings get placed in the root dist directory.
                if isinstance(datafile,basestring):
                    datafile = ("",[datafile])
                (dst,sources) = datafile
                if os.path.isabs(dst):
                    err = "cant freeze absolute data_file paths (%s)"
                    err = err % (dst,)
                    raise ValueError(err)
                dst = convert_path(dst)
                for src in sources:
                    src = convert_path(src)
                    yield (src,os.path.join(dst,os.path.basename(src)))
 
    def get_package_data(self):
        """Yield (source,destination) tuples for package data files.

        This method generates the names of all package data files to be
        included in the frozen app.  They should be placed in the library.zip
        or equivalent, alongside the python files for that package.
        """
        if self.distribution.package_data:
            for pkg,data in self.distribution.package_data.iteritems():
                pkg_dir = self.get_package_dir(pkg)
                pkg_path = pkg.replace(".","/")
                if isinstance(data,basestring):
                    data = [data]
                for dpattern in data:
                    dfiles = glob(os.path.join(pkg_dir,convert_path(dpattern)))
                    for nm in dfiles:
                        arcnm = pkg_path + nm[len(pkg_dir):]
                        yield (nm,arcnm)

    def get_package_dir(self,pkg):
        """Return directory where the given package is located.

        This was largely swiped from distutils, with some cleanups.
        """
        inpath = pkg.split(".")
        outpath = []
        if not self.distribution.package_dir:
            outpath = inpath
        else:
            while inpath:
                try:
                    dir = self.distribution.package_dir[".".join(inpath)]
                except KeyError:
                    outpath.insert(0, inpath[-1])
                    del inpath[-1]
                else:
                    outpath.insert(0, dir)
                    break
            else:
                try:
                    dir = self.package_dir[""]
                except KeyError:
                    pass
                else:
                    outpath.insert(0, dir)
        if outpath:
            return os.path.join(*outpath)
        else:
            return ""

    @staticmethod
    def get_msvcrt_private_assembly_files():
        """Get (source,destination) tuples for the MSVCRT DLLs, manifest etc.

        This method generates data_files tuples for the MSVCRT DLLs, manifest
        and associated paraphernalia.  Including these files is required for
        newer Python versions if you want to run on machines that don't have
        the latest C runtime installed *and* you don't want to run the special
        "vcredist_x86.exe" program during your installation process.

        Bundling is only performed on win32 paltforms, and only if you enable
        it explicitly.  Before doing so, carefully check whether you have a
        license to distribute these files.
        """
        cls = bdist_esky
        msvcrt_info = cls._get_msvcrt_info()
        if msvcrt_info is not None:
            msvcrt_name = msvcrt_info[0]
            #  Find installed manifest file with matching info
            for candidate in cls._find_msvcrt_manifest_files(msvcrt_name):
                manifest_file, msvcrt_dir = candidate
                try:
                    with open(manifest_file,"rb") as mf:
                        manifest_data = mf.read()
                        for info in msvcrt_info:
                            if info.encode() not in manifest_data:
                                break
                        else:
                            break
                except EnvironmentError:
                    pass
            else:
                err = "manifest for %s not found" % (msvcrt_info,)
                raise RuntimeError(err)
            #  Copy the manifest and matching directory into the freeze dir.
            manifest_name = msvcrt_name + ".manifest"
            yield (manifest_file,os.path.join(msvcrt_name,manifest_name))
            for fnm in os.listdir(msvcrt_dir):
                yield (os.path.join(msvcrt_dir,fnm),
                       os.path.join(msvcrt_name,fnm))

    @staticmethod
    def _get_msvcrt_info():
        """Get info about the MSVCRT in use by this python executable.

        This parses the name, version and public key token out of the exe
        manifest and returns them as a tuple.
        """
        try:
            manifest_str = winres.get_app_manifest()
        except EnvironmentError:
            return None
        manifest = minidom.parseString(manifest_str)
        for assembly in manifest.getElementsByTagName("assemblyIdentity"):
            name = assembly.attributes["name"].value
            if name.startswith("Microsoft") and name.endswith("CRT"):
                version = assembly.attributes["version"].value 
                pubkey = assembly.attributes["publicKeyToken"].value 
                return (name,version,pubkey)
        return None
        
    
    @staticmethod
    def _find_msvcrt_manifest_files(name):
        """Search the system for candidate MSVCRT manifest files.

        This method yields (manifest_file,msvcrt_dir) tuples giving a candidate
        manifest file for the given assembly name, and the directory in which
        the actual assembly data files are found.
        """
        cls = bdist_esky
        #  Search for redist files in a Visual Studio install
        progfiles = os.path.expandvars("%PROGRAMFILES%")
        for dnm in os.listdir(progfiles):
            if dnm.lower().startswith("microsoft visual studio"):
                dpath = os.path.join(progfiles,dnm,"VC","redist")
                for (subdir,_,filenames) in os.walk(dpath):
                    for fnm in filenames:
                        if name.lower() in fnm.lower():
                            if fnm.lower().endswith(".manifest"):
                                mf = os.path.join(subdir,fnm)
                                md = cls._find_msvcrt_dir_for_manifest(name,mf)
                                if md is not None:
                                    yield (mf,md)
        #  Search for manifests installed in the WinSxS directory
        winsxs_m = os.path.expandvars("%WINDIR%\\WinSxS\\Manifests")
        for fnm in os.listdir(winsxs_m):
            if name.lower() in fnm.lower():
                if fnm.lower().endswith(".manifest"):
                    mf = os.path.join(winsxs_m,fnm)
                    md = cls._find_msvcrt_dir_for_manifest(name,mf)
                    if md is not None:
                        yield (mf,md)
        winsxs = os.path.expandvars("%WINDIR%\\WinSxS")
        for fnm in os.listdir(winsxs):
            if name.lower() in fnm.lower():
                if fnm.lower().endswith(".manifest"):
                    mf = os.path.join(winsxs,fnm)
                    md = cls._find_msvcrt_dir_for_manifest(name,mf)
                    if md is not None:
                        yield (mf,md)

    @staticmethod
    def _find_msvcrt_dir_for_manifest(msvcrt_name,manifest_file):
        """Find the directory containing data files for the given manifest.

        This searches a few common locations for the data files that go with
        the given manifest file.  If a suitable directory is found then it is
        returned, otherwise None is returned.
        """
        #  The manifest file might be next to the dir, inside the dir, or
        #  in a subdir named "Manifests".  Walk around till we find it.
        msvcrt_dir = ".".join(manifest_file.split(".")[:-1])
        if os.path.isdir(msvcrt_dir):
            return msvcrt_dir
        msvcrt_basename = os.path.basename(msvcrt_dir)
        msvcrt_parent = os.path.dirname(os.path.dirname(msvcrt_dir))
        msvcrt_dir = os.path.join(msvcrt_parent,msvcrt_basename)
        if os.path.isdir(msvcrt_dir):
            return msvcrt_dir
        msvcrt_dir = os.path.join(msvcrt_parent,msvcrt_name)
        if os.path.isdir(msvcrt_dir):
            return msvcrt_dir
        return None


    def compile_to_bootstrap_exe(self,exe,source,relpath=None):
        """Compile the given sourcecode into a bootstrapping exe.

        This method compiles the given sourcecode into a stand-alone exe using
        PyPy, then stores that in the bootstrap env under the name of the given
        Executable object.  If the source has been previously compiled then a
        cached version of the exe may be used.
        """
        if not relpath:
            relpath = exe.name
        source = "__rpython__ = True\n" + source
        cdir = os.path.join(self.tempdir,"compile")
        if not os.path.exists(cdir):
            os.mkdir(cdir)
        source_hash = hashlib.md5(source).hexdigest()
        outname = "bootstrap_%s.%s" % (source_hash,get_platform())
        if exe.gui_only:
            outname += ".gui"
        if sys.platform == "win32":
            outname += ".exe"
        #  First try to use a precompiled version.
        if COMPILED_BOOTSTRAP_CACHE is not None:
            outfile = os.path.join(COMPILED_BOOTSTRAP_CACHE,outname)
            if os.path.exists(outfile):
                return self.copy_to_bootstrap_env(outfile,relpath)
        #  Otherwise we have to compile it anew.
        try:
            outfile = self._compiled_exes[(source_hash,exe.gui_only)]
        except KeyError:
            infile = os.path.join(cdir,"bootstrap.py")
            outfile = os.path.join(cdir,outname)
            with open(infile,"wt") as f:
                f.write(source)
            opts = dict(gui_only=exe.gui_only)
            pypyc.compile_rpython(infile,outfile,**opts)
            self._compiled_exes[(source_hash,exe.gui_only)] = outfile
        #  Try to save the compiled exe for future use.
        if COMPILED_BOOTSTRAP_CACHE is not None:
            cachedfile = os.path.join(COMPILED_BOOTSTRAP_CACHE,outname)
            try:
                shutil.copy2(outfile,cachedfile)
            except EnvironmentError:
                pass
        return self.copy_to_bootstrap_env(outfile,relpath)

    def copy_to_bootstrap_env(self,src,dst=None):
        """Copy the named file into the bootstrap environment.

        The filename is also added to the bootstrap manifest.
        """
        if dst is None:
            dst = src
        srcpath = os.path.join(self.freeze_dir,src)
        dstpath = os.path.join(self.bootstrap_dir,dst)
        if os.path.isdir(srcpath):
            self.copy_tree(srcpath,dstpath)
        else:
            if not os.path.isdir(os.path.dirname(dstpath)):
               self.mkpath(os.path.dirname(dstpath))
            self.copy_file(srcpath,dstpath)
        self.add_to_bootstrap_manifest(dstpath)
        return dstpath

    def add_to_bootstrap_manifest(self,dstpath):
        if not os.path.isdir(os.path.join(self.freeze_dir,ESKY_CONTROL_DIR)):
            os.mkdir(os.path.join(self.freeze_dir,ESKY_CONTROL_DIR))
        f_manifest = os.path.join(self.freeze_dir,ESKY_CONTROL_DIR,"bootstrap-manifest.txt")
        with open(f_manifest,"at") as f_manifest:
            f_manifest.seek(0,os.SEEK_END)
            if os.path.isdir(dstpath):
                for (dirnm,_,filenms) in os.walk(dstpath):
                    for fnm in filenms:
                        fpath = os.path.join(dirnm,fnm)
                        dpath = fpath[len(self.bootstrap_dir)+1:]
                        if os.sep != "/":
                            dpath = dpath.replace(os.sep,"/")
                        f_manifest.write(dpath)
                        f_manifest.write("\n")
            else:
                dst = dstpath[len(self.bootstrap_dir)+1:]
                if os.sep != "/":
                    dst = dst.replace(os.sep,"/")
                f_manifest.write(dst)
                f_manifest.write("\n")


class bdist_esky_patch(Command):
    """Create a patch for a frozen application in 'esky' format.

    This distutils command can be used to create a patch file between two
    versions of an application frozen with esky.  Such a patch can be used
    for differential updates between application versions.
    """

    user_options = [
                    ('dist-dir=', 'd',
                     "directory to put final built distributions in"),
                    ('from-version=', None,
                     "version against which to produce patch"),
                   ]

    def initialize_options(self):
        self.dist_dir = None
        self.from_version = None

    def finalize_options(self):
        self.set_undefined_options('bdist',('dist_dir', 'dist_dir'))

    def run(self):
        fullname = self.distribution.get_fullname()
        platform = get_platform()
        vdir = "%s.%s" % (fullname,platform,)
        appname = split_app_version(vdir)[0]
        #  Ensure we have current version's esky, as target for patch.
        target_esky = os.path.join(self.dist_dir,vdir+".zip")
        if not os.path.exists(target_esky):
            self.run_command("bdist_esky")
        #  Generate list of source eskys to patch against.
        if self.from_version:
            source_vdir = join_app_version(appname,self.from_version,platform)
            source_eskys = [os.path.join(self.dist_dir,source_vdir+".zip")]
        else:
            source_eskys = []
            for nm in os.listdir(self.dist_dir):
                if target_esky.endswith(nm):
                    continue
                if nm.startswith(appname+"-") and nm.endswith(platform+".zip"):
                    source_eskys.append(os.path.join(self.dist_dir,nm))
        #  Write each patch, transparently unzipping the esky
        for source_esky in source_eskys:
            target_vdir = os.path.basename(source_esky)[:-4]
            target_version = split_app_version(target_vdir)[1]
            patchfile = vdir+".from-%s.patch" % (target_version,)
            patchfile = os.path.join(self.dist_dir,patchfile)
            print "patching", target_esky, "against", source_esky, "=>", patchfile
            if not self.dry_run:
                try:
                    esky.patch.main(["-Z","diff",source_esky,target_esky,patchfile])
                except:
                    import traceback
                    traceback.print_exc()
                    raise


#  Monkey-patch distutils to include our commands by default.
distutils.command.__all__.append("bdist_esky")
distutils.command.__all__.append("bdist_esky_patch")
sys.modules["distutils.command.bdist_esky"] = sys.modules["esky.bdist_esky"]
sys.modules["distutils.command.bdist_esky_patch"] = sys.modules["esky.bdist_esky"]






"""

  esky.bdist_esky.pypy_libpython:  load python DLL into pypy bootstrap exe


This module provides the class "libpython", an RPython-compatible class for
loading and exposing a python environment using clibffi.  It's used by the
pypy-compiled bootstrap exes to bootstrap a version dir in-process.

"""


from pypy.rlib import clibffi
from pypy.rpython.lltypesystem import rffi, lltype


class libpython(object):

    file_input = 257


    def __init__(self,library_path):
        self.lib = clibffi.CDLL(library_path)
        self._libc = clibffi.CDLL(clibffi.get_libc_name())


    def Set_NoSiteFlag(self,value):
        addr = self.lib.getaddressindll("Py_NoSiteFlag")
        memset = self._libc.getpointer("memset",[clibffi.ffi_type_pointer,clibffi.ffi_type_uint,clibffi.ffi_type_uint],clibffi.ffi_type_void)
        memset.push_arg(addr)
        memset.push_arg(value)
        memset.push_arg(1)
        memset.call(lltype.Void)


    def Set_FrozenFlag(self,value):
        addr = self.lib.getaddressindll("Py_FrozenFlag")
        memset = self._libc.getpointer("memset",[clibffi.ffi_type_pointer,clibffi.ffi_type_uint,clibffi.ffi_type_uint],clibffi.ffi_type_void)
        memset.push_arg(addr)
        memset.push_arg(value)
        memset.push_arg(1)
        memset.call(lltype.Void)


    def Set_IgnoreEnvironmentFlag(self,value):
        addr = self.lib.getaddressindll("Py_IgnoreEnvironmentFlag")
        memset = self._libc.getpointer("memset",[clibffi.ffi_type_pointer,clibffi.ffi_type_uint,clibffi.ffi_type_uint],clibffi.ffi_type_void)
        memset.push_arg(addr)
        memset.push_arg(value)
        memset.push_arg(1)
        memset.call(lltype.Void)


    def Set_OptimizeFlag(self,value):
        addr = self.lib.getaddressindll("Py_OptimizeFlag")
        memset = self._libc.getpointer("memset",[clibffi.ffi_type_pointer,clibffi.ffi_type_uint,clibffi.ffi_type_uint],clibffi.ffi_type_void)
        memset.push_arg(addr)
        memset.push_arg(value)
        memset.push_arg(1)
        memset.call(lltype.Void)


    def Initialize(self):
        impl = self.lib.getpointer("Py_Initialize",[],clibffi.ffi_type_void)
        impl.call(lltype.Void)


    def Finalize(self):
        impl = self.lib.getpointer("Py_Finalize",[],clibffi.ffi_type_void)
        impl.call(lltype.Void)


    def Err_Occurred(self):
        impl = self.lib.getpointer("PyErr_Occurred",[],clibffi.ffi_type_pointer)
        return impl.call(rffi.VOIDP)


    def Err_Print(self):
        impl = self.lib.getpointer("PyErr_Print",[],clibffi.ffi_type_void)
        impl.call(lltype.Void)


    def _error(self):
        err = self.Err_Occurred()
        if err:
            self.Err_Print()
            raise RuntimeError("an error occurred")


    def Run_SimpleString(self,string):
        impl = self.lib.getpointer("PyRun_SimpleString",[clibffi.ffi_type_pointer],clibffi.ffi_type_sint)
        buf = rffi.str2charp(string)
        impl.push_arg(rffi.cast(rffi.VOIDP,buf))
        res = impl.call(rffi.INT)
        rffi.free_charp(buf)
        if res < 0:
            self._error()


    def Run_String(self,string,start,globals=None,locals=None):
        if globals is None:
            globals = 0
        if locals is None:
            locals = 0
        impl = self.lib.getpointer("PyRun_String",[clibffi.ffi_type_pointer,clibffi.ffi_type_sint,clibffi.ffi_type_pointer,clibffi.ffi_type_pointer],clibffi.ffi_type_pointer)
        buf = rffi.str2charp(string)
        impl.push_arg(rffi.cast(rffi.VOIDP,buf))
        impl.push_arg(start)
        impl.push_arg(globals)
        impl.push_arg(locals)
        res = impl.call(rffi.VOIDP)
        rffi.free_charp(buf)
        if not res:
            self._error()
        return res


    def GetProgramFullPath(self):
        impl = self.lib.getpointer("Py_GetProgramFullPath",[],clibffi.ffi_type_pointer)
        return rffi.charp2str(impl.call(rffi.CCHARP))


    def SetPythonHome(self,path):
        return
        impl = self.lib.getpointer("Py_SetPythonHome",[clibffi.ffi_type_pointer],clibffi.ffi_type_void)
        buf = rffi.str2charp(path)
        impl.push_arg(buf)
        impl.call(lltype.Void)
        rffi.free_charp(buf)


    # TODO: this seems to cause type errors during building
    def Sys_SetArgv(self,argv):
        impl = self.lib.getpointer("PySys_SetArgv",[clibffi.ffi_type_sint,clibffi.ffi_type_pointer],clibffi.ffi_type_void)
        impl.push_arg(len(argv))
        buf = rffi.liststr2charpp(argv)
        impl.push_arg(rffi.cast(rffi.VOIDP,buf))
        impl.call(lltype.Void)
        rffi.free_charpp(buf)


    def Sys_SetPath(self,path):
        impl = self.lib.getpointer("PySys_SetPath",[clibffi.ffi_type_pointer],clibffi.ffi_type_void)
        buf = rffi.str2charp(path)
        impl.push_arg(rffi.cast(rffi.VOIDP,buf))
        impl.call(lltype.Void)
        rffi.free_charp(buf)


    def Eval_GetBuiltins(self):
        impl = self.lib.getpointer("PyEval_GetBuiltins",[],clibffi.ffi_type_pointer)
        d = impl.call(rffi.VOIDP)
        if not d:
            self._error()
        return d


    def Import_ImportModule(self,name):
        impl = self.lib.getpointer("PyImport_ImportModule",[clibffi.ffi_type_pointer],clibffi.ffi_type_pointer)
        buf = rffi.str2charp(name)
        impl.push_arg(rffi.cast(rffi.VOIDP,buf))
        mod = impl.call(rffi.VOIDP)
        rffi.free_charp(buf)
        if not mod:
            self._error()
        return mod


    def Object_GetAttr(self,obj,attr):
        impl = self.lib.getpointer("PyObject_GetAttr",[clibffi.ffi_type_pointer,clibffi.ffi_type_pointer],clibffi.ffi_type_pointer)
        impl.push_arg(obj)
        impl.push_arg(attr)
        a = impl.call(rffi.VOIDP)
        if not a:
            self._error()
        return a


    def Object_GetAttrString(self,obj,attr):
        impl = self.lib.getpointer("PyObject_GetAttrString",[clibffi.ffi_type_pointer,clibffi.ffi_type_pointer],clibffi.ffi_type_pointer)
        impl.push_arg(obj)
        buf = rffi.str2charp(attr)
        impl.push_arg(rffi.cast(rffi.VOIDP,buf))
        a = impl.call(rffi.VOIDP)
        rffi.free_charp(buf)
        if not a:
            self._error()
        return a


    def Object_SetAttr(self,obj,attr,val):
        impl = self.lib.getpointer("PyObject_SetAttr",[clibffi.ffi_type_pointer,clibffi.ffi_type_pointer,clibffi.ffi_type_pointer],clibffi.ffi_type_sint)
        impl.push_arg(obj)
        impl.push_arg(attr)
        impl.push_arg(val)
        res = impl.call(rffi.INT)
        if res < 0:
            self._error()
        return None


    def Object_SetAttrString(self,obj,attr,val):
        impl = self.lib.getpointer("PyObject_SetAttrString",[clibffi.ffi_type_pointer,clibffi.ffi_type_pointer,clibffi.ffi_type_pointer],clibffi.ffi_type_sint)
        impl.push_arg(obj)
        buf = rffi.str2charp(attr)
        impl.push_arg(rffi.cast(rffi.VOIDP,buf))
        impl.push_arg(val)
        res = impl.call(rffi.INT)
        rffi.free_charp(buf)
        if res < 0:
            self._error()
        return None


    def Dict_New(self):
        impl = self.lib.getpointer("PyDict_New",[],clibffi.ffi_type_pointer)
        d = impl.call(rffi.VOIDP)
        if not d:
            self._error()
        return d


    def Dict_SetItemString(self,d,key,value):
        impl = self.lib.getpointer("PyDict_SetItemString",[clibffi.ffi_type_pointer,clibffi.ffi_type_pointer,clibffi.ffi_type_pointer],clibffi.ffi_type_sint)
        impl.push_arg(d)
        buf = rffi.str2charp(key)
        impl.push_arg(rffi.cast(rffi.VOIDP,buf))
        impl.push_arg(value)
        d = impl.call(rffi.INT)
        rffi.free_charp(buf)
        if d < 0:
            self._error()


    def List_New(self,size=0):
        impl = self.lib.getpointer("PyList_New",[clibffi.ffi_type_uint],clibffi.ffi_type_pointer)
        impl.push_arg(size)
        l = impl.call(rffi.VOIDP)
        if not l:
            self._error()
        return l


    def List_Size(self,l):
        impl = self.lib.getpointer("PyList_Size",[clibffi.ffi_type_pointer],clibffi.ffi_type_uint)
        impl.push_arg(l)
        s = impl.call(rffi.INT)
        if s < 0:
            self._error()
        return s


    def List_SetItem(self,l,i,v):
        impl = self.lib.getpointer("PyList_SetItem",[clibffi.ffi_type_pointer,clibffi.ffi_type_uint,clibffi.ffi_type_pointer],clibffi.ffi_type_sint)
        impl.push_arg(l)
        impl.push_arg(i)
        impl.push_arg(v)
        res = impl.call(rffi.INT)
        if res < 0:
            self._error()


    def List_Append(self,l,v):
        impl = self.lib.getpointer("PyList_Append",[clibffi.ffi_type_pointer,clibffi.ffi_type_pointer],clibffi.ffi_type_sint)
        impl.push_arg(l)
        impl.push_arg(v)
        res = impl.call(rffi.INT)
        if res < 0:
            self._error()


    def String_FromString(self,s):
        impl = self.lib.getpointer("PyString_FromString",[clibffi.ffi_type_pointer],clibffi.ffi_type_pointer)
        buf = rffi.str2charp(s)
        impl.push_arg(rffi.cast(rffi.VOIDP,buf))
        ps = impl.call(rffi.VOIDP)
        rffi.free_charp(buf)
        if not ps:
            self._error()
        return ps


    def String_FromStringAndSize(self,s,size):
        impl = self.lib.getpointer("PyString_FromStringAndSize",[clibffi.ffi_type_pointer,clibffi.ffi_type_uint],clibffi.ffi_type_pointer)
        if not s:
            buf = None
            impl.push_arg(None)
        else:
            buf = rffi.str2charp(s)
            impl.push_arg(rffi.cast(rffi.VOIDP,buf))
        impl.push_arg(size)
        ps = impl.call(rffi.VOIDP)
        if s:
            rffi.free_charp(buf)
        if not ps:
            self._error()
        return ps


    def String_AsString(self,s):
        impl = self.lib.getpointer("PyString_AsString",[clibffi.ffi_type_pointer],clibffi.ffi_type_pointer)
        impl.push_arg(s)
        buf = impl.call(rffi.VOIDP)
        if not buf:
            self._error()
        return buf





#  Copyright (c) 2009-2010, Cloud Matrix Pty. Ltd.
#  All rights reserved; available under the terms of the BSD License.
"""

  esky.bdist_esky.f_py2exe:  bdist_esky support for py2exe

"""

from __future__ import with_statement


import os
import re
import sys
import imp
import time
import zipfile
import marshal
import struct
import shutil
import inspect
import zipfile
import ctypes


from py2exe.build_exe import py2exe

import esky
from esky.util import is_core_dependency, ESKY_CONTROL_DIR
from esky import winres

try:
    import py2exe.mf as modulefinder
except ImportError:
    modulefinder = None

#  Hack to make win32com work seamlessly with py2exe
if modulefinder is not None:
  try:
    import win32com
    for p in win32com.__path__[1:]:
        modulefinder.AddPackagePath("win32com", p)
    for extra in ["win32com.shell"]: #,"win32com.mapi"
        __import__(extra)
        m = sys.modules[extra]
        for p in m.__path__[1:]:
           modulefinder.AddPackagePath(extra, p)
  except ImportError:
     pass


class custom_py2exe(py2exe): 
    """Custom py2exe command subclass.

    This py2exe command subclass incorporates some well-known py2exe "hacks"
    to make common third-party packages work better.
    """

    def create_modulefinder(self):
        mf = py2exe.create_modulefinder(self)
        self.__mf = mf
        return mf

    def build_manifest(self,target,template):
        (mfest,mid) = py2exe.build_manifest(self,target,template)
        #  Hack to get proper UI theme when freezing wxPython
        if mfest is not None:
            if "wx" in self.__mf.modules:
                mfest = mfest.replace("</assembly>","""
                    <dependency>
                      <dependentAssembly>
                        <assemblyIdentity
                         type="win32"
                         name="Microsoft.Windows.Common-Controls"
                         version="6.0.0.0"
                         processorArchitecture="*"
                         publicKeyToken="6595b64144ccf1df"
                         language="*" />
                      </dependentAssembly>
                   </dependency>
                 </assembly>""")
        return (mfest,mid)


def freeze(dist):
    """Freeze the given distribution data using py2exe."""
    includes = dist.includes
    excludes = dist.excludes
    options = dist.freezer_options
    #  Merge in any encludes/excludes given in freezer_options
    includes.append("esky")
    for inc in options.pop("includes",()):
        includes.append(inc)
    for exc in options.pop("excludes",()):
        excludes.append(exc)
    if "pypy" not in includes and "pypy" not in excludes:
        excludes.append("pypy")
    #  py2exe expects some arguments on the main distribution object.
    #  We handle data_files ourselves, so fake it out for py2exe.
    if getattr(dist.distribution,"console",None):
        msg = "don't call setup(console=[...]) with esky;"
        msg += " use setup(scripts=[...]) instead"
        raise RuntimeError(msg)
    if getattr(dist.distribution,"windows",None):
        msg = "don't call setup(windows=[...]) with esky;"
        msg += " use setup(scripts=[...]) instead"
        raise RuntimeError(msg)
    dist.distribution.console = []
    dist.distribution.windows = []
    my_data_files = dist.distribution.data_files
    dist.distribution.data_files = []
    for exe in dist.get_executables():
        #  Pass any executable kwds through to py2exe.
        #  We handle "icon" and "gui_only" ourselves.
        s = exe._kwds.copy()
        s["script"] = exe.script
        s["dest_base"] = exe.name[:-4]
        if exe.icon is not None and "icon_resources" not in s:
            s["icon_resources"] = [(1,exe.icon)]
        if exe.gui_only:
            dist.distribution.windows.append(s)
        else:
            dist.distribution.console.append(s)
    if "zipfile" in options:
        dist.distribution.zipfile = options.pop("zipfile")
    #  Create the py2exe cmd and adjust its options
    cmd = custom_py2exe(dist.distribution)
    cmd.includes = includes
    cmd.excludes = excludes
    if "bundle_files" in options:
        if options["bundle_files"] < 3 and dist.compile_bootstrap_exes:
             err = "can't compile bootstrap exes when bundle_files < 3"
             raise RuntimeError(err)
    for (nm,val) in options.iteritems():
        setattr(cmd,nm,val)
    cmd.dist_dir = dist.freeze_dir
    cmd.finalize_options()
    #  Actually run the freeze process
    cmd.run()
    #  Copy data files into the freeze dir
    dist.distribution.data_files = my_data_files
    for (src,dst) in dist.get_data_files():
        dst = os.path.join(dist.freeze_dir,dst)
        dstdir = os.path.dirname(dst)
        if not os.path.isdir(dstdir):
            dist.mkpath(dstdir)
        dist.copy_file(src,dst)
    #  Place a marker file so we know how it was frozen
    os.mkdir(os.path.join(dist.freeze_dir,ESKY_CONTROL_DIR))
    marker_file = os.path.join(ESKY_CONTROL_DIR,"f-py2exe-%d%d.txt")%sys.version_info[:2]
    open(os.path.join(dist.freeze_dir,marker_file),"w").close()
    #  Copy package data into the library.zip
    #  For now, we don't try to put package data into a bundled zipfile.
    dist_zipfile = dist.distribution.zipfile
    if dist_zipfile is None:
        for (src,arcnm) in dist.get_package_data():
            err = "zipfile=None can't be used with package_data (yet...)"
            raise RuntimeError(err)
    elif not cmd.skip_archive:
        lib = zipfile.ZipFile(os.path.join(dist.freeze_dir,dist_zipfile),"a")
        for (src,arcnm) in dist.get_package_data():
            lib.write(src,arcnm)
        lib.close()
    else:
        for (src,arcnm) in dist.get_package_data():
            lib = os.path.join(dist.freeze_dir,os.path.dirname(dist_zipfile))
            dest = os.path.join(lib, os.path.dirname(src))
            f = os.path.basename(src)
            if not os.path.isdir(dest):
                dist.mkpath(dest)
            dist.copy_file(src,os.path.join(dest, f))
    #  There's no need to copy library.zip into the bootstrap env, as the
    #  chainloader will run before py2exe goes looking for it.
    pass
    #  Create the bootstraping code, using custom code if specified.
    #  It gets stored as a marshalled list of code objects directly in the exe.
    esky_name = re.escape(dist.distribution.get_name())
    code_source = ["__esky_name__ = '%s'" % (esky_name,)]
    code_source.append(inspect.getsource(esky.bootstrap))
    if dist.compile_bootstrap_exes:
        from esky.bdist_esky import pypy_libpython
        from esky.bdist_esky import pypy_winres
        code_source.append(inspect.getsource(pypy_libpython))
        code_source.append(inspect.getsource(pypy_winres))
        code_source.append(_CUSTOM_PYPY_CHAINLOADER)
        code_source.append(dist.get_bootstrap_code())
        code_source = "\n".join(code_source)
        for exe in dist.get_executables(normalise=False):
            if not exe.include_in_bootstrap_env:
                continue
            fexe = os.path.join(dist.freeze_dir,exe.name)
            bsexe = dist.compile_to_bootstrap_exe(exe,code_source)
            winres.copy_safe_resources(fexe,bsexe)
        #  We may also need the bundled MSVCRT libs
        for nm in os.listdir(dist.freeze_dir):
            if is_core_dependency(nm) and nm.startswith("Microsoft"):
                dist.copy_to_bootstrap_env(nm)
    else:
        code_source.append(_FAKE_ESKY_BOOTSTRAP_MODULE)
        code_source.append(_CUSTOM_WIN32_CHAINLOADER)
        code_source.append(dist.get_bootstrap_code())
        code_source.append("bootstrap()")
        code_source = "\n".join(code_source)
        code = marshal.dumps([compile(code_source,"__main__.py","exec")])
        coderes = struct.pack("iiii",
                         0x78563412, # magic value used for integrity checking,
                         0, # no optimization
                         False,  # normal buffered output
                         len(code),
                         ) + "\x00" + code + "\x00\x00"
        #  Copy any core dependencies into the bootstrap env.
        for nm in os.listdir(dist.freeze_dir):
            if is_core_dependency(nm):
                dist.copy_to_bootstrap_env(nm)
        #  Copy the loader program for each script into the bootstrap env.
        for exe in dist.get_executables(normalise=False):
            if not exe.include_in_bootstrap_env:
                continue
            exepath = dist.copy_to_bootstrap_env(exe.name)
            #  Insert the bootstrap code into the exe as a resource.
            #  This appears to have the happy side-effect of stripping any
            #  extra data from the end of the exe, which is exactly what we
            #  want when zipfile=None is specified; otherwise each bootstrap
            #  exe would also contain the whole bundled zipfile.
            winres.add_resource(exepath,coderes,u"PYTHONSCRIPT",1,0)
        #  If the python dll hasn't been copied into the bootstrap env,
        #  make sure it's stored in each bootstrap dll as a resource.
        pydll = u"python%d%d.dll" % sys.version_info[:2]
        if not os.path.exists(os.path.join(dist.bootstrap_dir,pydll)):
            buf = ctypes.create_string_buffer(3000)
            GetModuleFileNameA = ctypes.windll.kernel32.GetModuleFileNameA
            if not GetModuleFileNameA(sys.dllhandle,ctypes.byref(buf),3000):
                raise ctypes.WinError()
            with open(buf.value,"rb") as f:
                pydll_bytes = f.read()
            for exe in dist.get_executables(normalise=False):
                if not exe.include_in_bootstrap_env:
                    continue
                exepath = os.path.join(dist.bootstrap_dir,exe.name)
                try:
                    winres.load_resource(exepath,pydll.upper(),1,0)
                except EnvironmentError:
                    winres.add_resource(exepath,pydll_bytes,pydll.upper(),1,0)

#  Code to fake out any bootstrappers that try to import from esky.
_FAKE_ESKY_BOOTSTRAP_MODULE = """
class __fake:
  __all__ = ()
sys.modules["esky"] = __fake()
sys.modules["esky.bootstrap"] = __fake()
"""


#  On Windows, execv is flaky and expensive.  If the chainloader is the same
#  python version as the target exe, we can munge sys.path to bootstrap it
#  into the existing process.
#
#  We need to read the script to execute as a resource from the exe, so this
#  only works if we can bootstrap a working ctypes module.  We then insert
#  the source code from esky.winres.load_resource directly into this function.
#
_CUSTOM_WIN32_CHAINLOADER = """

_orig_chainload = _chainload

def _chainload(target_dir):
  # careful to escape percent-sign, this gets interpolated below
  marker_file = pathjoin(ESKY_CONTROL_DIR,"f-py2exe-%%d%%d.txt")%%sys.version_info[:2]
  pydll = "python%%s%%s.dll" %% sys.version_info[:2]
  mydir = dirname(sys.executable)
  if not exists(pathjoin(target_dir,marker_file)):
      return _orig_chainload(target_dir)
  for nm in listdir(target_dir):
      if nm == pydll:
          continue
      if nm.lower().startswith("msvcr"):
          continue
      if nm.lower().endswith(".pyd") or nm.lower().endswith(".dll"):
          #  The freeze dir contains unbundled C extensions.
          #  Since they're linked against a physical python DLL, we
          #  can't chainload them unless we have one too.
          if not exists(pathjoin(mydir,pydll)):
              return _orig_chainload(target_dir)
          else:
               break
  # munge the environment to pretend we're in the target dir
  try:
      import nt
  except ImportError:
      return _orig_chainload(target_dir)
  sys.bootstrap_executable = sys.executable
  sys.executable = pathjoin(target_dir,basename(sys.executable))
  verify(sys.executable)
  sys.prefix = sys.prefix.replace(mydir,target_dir)
  sys.argv[0] = sys.executable
  for i in xrange(len(sys.path)):
      sys.path[i] = sys.path[i].replace(mydir,target_dir)
  curdir = getcwd()
  newdir = curdir.replace(mydir,target_dir)
  if newdir != curdir:
      nt.chdir(newdir)
  libfile = pathjoin(target_dir,"library.zip")
  if libfile not in sys.path:
      if exists(libfile):
          sys.path.append(libfile)
      else:
          sys.path.append(target_dir)
  # try to import the modules we need for bootstrapping
  try:
      import zipextimporter; zipextimporter.install()
  except ImportError:
      pass
  try:
      import ctypes
      import struct
      import marshal
      import msvcrt
  except ImportError:
      return _orig_chainload(target_dir)
  # the source for esky.winres.load_resource gets inserted below:
  from ctypes import c_char, POINTER
  k32 = ctypes.windll.kernel32
  LOAD_LIBRARY_AS_DATAFILE = 0x00000002
  _DEFAULT_RESLANG = 1033
  %s
  # now we magically have the load_resource function :-)
  try:
      data = load_resource(sys.executable,u"PYTHONSCRIPT",1,0)
  except EnvironmentError:
      #  This will trigger if sys.executable doesn't exist.
      #  Falling back to the original chainloader will account for
      #  the unlikely case where sys.executable is a backup file.
      return _orig_chainload(target_dir)
  else:
      sys.modules.pop("esky",None)
      sys.modules.pop("esky.bootstrap",None)
      headsz = struct.calcsize("iiii")
      (magic,optmz,unbfrd,codesz) = struct.unpack("iiii",data[:headsz])
      assert magic == 0x78563412
      # Set up the environment requested by "optimized" and "unbuffered"
      try:
          opt_var = ctypes.c_int.in_dll(ctypes.pythonapi,"Py_OptimizeFlag")
          opt_var.value = optmz
      except ValueError:
          pass
      if unbfrd:
          msvcrt.setmode(0,nt.O_BINARY)
          msvcrt.setmode(1,nt.O_BINARY)
          ctypes.pythonapi.PyFile_AsFile.argtypes = (ctypes.py_object,)
          if hasattr(ctypes.cdll.msvcrt,"setvbuf"):
              def setunbuf(f):
                  fp = ctypes.pythonapi.PyFile_AsFile(f)
                  ctypes.cdll.msvcrt.setvbuf(fp,None,4,512)
          else:
              def setunbuf(fd,mode):
                  fp = ctypes.pythonapi.PyFile_AsFile(f)
                  ctypes.cdll.msvcrt.setbuf(fp,None)
          setunbuf(sys.stdin)
          setunbuf(sys.stdout)
          setunbuf(sys.stderr)
      # skip over the archive name to find start of code
      codestart = headsz
      while data[codestart] != "\\0":
          codestart += 1
      codestart += 1
      codeend = codestart + codesz
      codelist = marshal.loads(data[codestart:codeend])
      # Execute all code in the context of __main__ module.
      d_locals = d_globals = sys.modules["__main__"].__dict__
      d_locals["__name__"] = "__main__"
      for code in codelist:
          exec code in d_globals, d_locals
      raise SystemExit(0)
""" % (inspect.getsource(winres.load_resource).replace("\n","\n"+" "*4),)


#  On Windows, execv is flaky and expensive.  Since the pypy-compiled bootstrap
#  exe doesn't have a python runtime, it needs to chainload the one from the
#  target version dir before trying to bootstrap in-process.
_CUSTOM_PYPY_CHAINLOADER = """

import nt
from pypy.rlib.rstruct.runpack import runpack

import time;

_orig_chainload = _chainload

def _chainload(target_dir):
  mydir = dirname(sys.executable)
  pydll = pathjoin(target_dir,"python%s%s.dll" % sys.version_info[:2])
  if not exists(pydll):
      return _orig_chainload(target_dir)
  else:

      #  Munge the environment for DLL loading purposes
      try:
          environ["PATH"] = environ["PATH"] + ";" + target_dir
      except KeyError:
          environ["PATH"] = target_dir

      #  Get the target python env up and running
      verify(pydll)
      py = libpython(pydll)
      py.Set_NoSiteFlag(1)
      py.Set_FrozenFlag(1)
      py.Set_IgnoreEnvironmentFlag(1)
      py.SetPythonHome("")
      py.Initialize()

      #  Extract the marshalled code data from the target executable,
      #  store it into a python string object.
      target_exe = pathjoin(target_dir,basename(sys.executable))
      verify(target_exe)
      try:
          py_data = load_resource_pystr(py,target_exe,"PYTHONSCRIPT",1,0)
      except EnvironmentError:
          return _orig_chainload(target_dir)
      data = py.String_AsString(py_data)
      headsz = 16  # <-- struct.calcsize("iiii")
      headdata = rffi.charpsize2str(rffi.cast(rffi.CCHARP,data),headsz)
      (magic,optmz,unbfrd,codesz) = runpack("iiii",headdata)
      assert magic == 0x78563412
      # skip over the archive name to find start of code
      codestart = headsz
      while data[codestart] != "\\0":
          codestart += 1
      codestart += 1
      codeend = codestart + codesz
      assert codeend > 0

      #  Tweak the python env according to the py2exe frozen metadata
      py.Set_OptimizeFlag(optmz)
      # TODO: set up buffering
      #if unbfrd:
      #     setmode(0,nt.O_BINARY)
      #     setmode(1,nt.O_BINARY)
      #     setvbuf(stdin,NULL,4,512)
      #     setvbuf(stdout,NULL,4,512)
      #     setvbuf(stderr,NULL,4,512)

      #  Preted the python env is running from within the frozen executable
      syspath = "%s;%s\\library.zip;%s" % (target_exe,target_dir,target_dir,)
      py.Sys_SetPath(syspath);
      sysmod = py.Import_ImportModule("sys")
      sysargv = py.List_New(len(sys.argv))
      for i in xrange(len(sys.argv)):
          py.List_SetItem(sysargv,i,py.String_FromString(sys.argv[i]))
      py.Object_SetAttrString(sysmod,"argv",sysargv)
      py.Object_SetAttrString(sysmod,"frozen",py.String_FromString("py2exe"))
      py.Object_SetAttrString(sysmod,"executable",py.String_FromString(target_exe))
      py.Object_SetAttrString(sysmod,"bootstrap_executable",py.String_FromString(sys.executable))
      py.Object_SetAttrString(sysmod,"prefix",py.String_FromString(dirname(target_exe)))

      curdir = getcwd()
      if curdir == mydir:
          nt.chdir(target_dir)

      #  Execute the marshalled list of code objects
      globals = py.Dict_New()
      py.Dict_SetItemString(globals,"__builtins__",py.Eval_GetBuiltins())
      py.Dict_SetItemString(globals,"FROZEN_DATA",py_data)
      runcode =  "FROZEN_DATA = FROZEN_DATA[%d:%d]\\n" % (codestart,codeend,)
      runcode +=  "import sys\\n"
      runcode +=  "import marshal\\n"
      runcode += "d_locals = d_globals = sys.modules['__main__'].__dict__\\n"
      runcode += "d_locals['__name__'] = '__main__'\\n"
      runcode += "for code in marshal.loads(FROZEN_DATA):\\n"
      runcode += "  exec code in d_globals, d_locals\\n"
      py.Run_String(runcode,py.file_input,globals)

      #  Clean up after execution.
      py.Finalize()
      sys.exit(0)

"""


